#define CORE_NAME "eth_rcm2"

#include <autoconf.h>
#if defined(CONFIG_ETHERNET_DEBUG)
#define DEBUG
#define VERBOSE_DEBUG
#define TRACE_DEBUG
#endif
#include <hermit.h>
#include <herrno.h>
#include <io.h>
#include <arch/pinmux.h>
#include <string.h>
#include <mac.h>
#include <net/net.h>
#include <net/eth.h>
#include <net/eth_util.h>
#include <delay.h>
#include <platform/pinmux.h>
#include <rcm2_gpio.h>

extern int rcm2_eth_request_port(void);

#define ETH_BASE_ADDR	0xee700200

#define ETH_EDMR	0x0000
#define 	EDMR_EL		0x40 /* Litte endian */
#define 	EMDR_DESC_R	0x30 /* Descriptor reserve size */
#define 	EMDR_DESC_64	0x20 /* Descriptor size is 64byte */
#define 	EDMR_DL1	0x20
#define 	EDMR_DL0	0x10
#define 	EDMR_SRST	0x01
#define ETH_EDTRR	0x0008
#define 	EDTRR_TRNS	0x01
#define ETH_EDRRR	0x0010
#define 	EDRRR_R		0x01
#define ETH_TDLAR	0x0018
#define ETH_RDLAR	0x0020
#define ETH_EESR	0x0028
#define ETH_EESIPR	0x0030
#define ETH_TRSCER	0x0038
#define ETH_RMFCR	0x0040
#define ETH_TFTR	0x0048
#define ETH_FDR		0x0050
#define 	FIFO_SIZE_T	0x00000700
#define 	FIFO_SIZE_R	0x00000007
#define ETH_RMCR	0x0058
#define 	RMCR_RST	0x01
#define ETH_TFUCR	0x0064
#define ETH_RFOCR	0x0068
#define ETH_FCFTR	0x0070
#define 	FCFTR_RFF2	0x00040000
#define 	FCFTR_RFF1	0x00020000
#define 	FCFTR_RFF0	0x00010000
#define 	FCFTR_RFD2	0x00000004
#define 	FCFTR_RFD1	0x00000002
#define 	FCFTR_RFD0	0x00000001
#define 	FIFO_F_D_RFF	(FCFTR_RFF2|FCFTR_RFF1|FCFTR_RFF0)
#define 	FIFO_F_D_RFD	(FCFTR_RFD2|FCFTR_RFD1|FCFTR_RFD0)
#define ETH_RPADIR	0x0078
#define ETH_TRIMD	0x007c
#define ETH_RBWAR	0x00c8
#define ETH_RDFAR	0x00cc
#define ETH_TBRAR	0x00d4
#define ETH_TDFAR	0x00d8

#define ETH_ECMR	0x0100
#define 	ECMR_ZPF	0x00080000
#define 	ECMR_PFR	0x00040000
#define 	ECMR_RXF	0x00020000
#define 	ECMR_TXF	0x00010000
#define 	ECMR_MCT	0x00002000
#define 	ECMR_PRCEF	0x00001000
#define 	ECMR_PMDE	0x00000200
#define 	ECMR_RE		0x00000040
#define 	ECMR_TE		0x00000020
#define 	ECMR_ILB	0x00000008
#define 	ECMR_ELB	0x00000004
#define 	ECMR_DM		0x00000002
#define 	ECMR_PRM	0x00000001
#define 	ECMR_RTM	0x00000004
#define 	ECMR_CHG_DM	(ECMR_ZPF | ECMR_PFR | ECMR_RXF | ECMR_TXF)

#define ETH_RFLR	0x0108
#define 	RFLR_RFL_MIN	0x05ee	/* Recv Frame length 1518 bytes */

#define ETH_ECSR	0x0110
#define ETH_ECSIPR	0x0118
#define ETH_PIR		0x0120
#define 	PIR_MDI		0x08
#define 	PIR_MDO		0x04
#define 	PIR_MMD		0x02
#define 	PIR_MDC		0x01
#define ETH_PSR		0x0128
#define ETH_RDMLR	0x0140
#define ETH_IPGR	0x0150
#define ETH_APR		0x0154
#define ETH_MPR		0x0158
#define ETH_TPAUSER	0x0164
#define ETH_RFCF	0x0160
#define ETH_TPAUSECR	0x0168
#define ETH_BCFRR	0x016c
#define ETH_MAHR	0x01c0
#define ETH_MALR	0x01c8
#define ETH_TROCR	0x01d0
#define ETH_CDCR	0x01d4
#define ETH_LCCR	0x01d8
#define ETH_CNDCR	0x01dc
#define ETH_CEFCR	0x01e4
#define ETH_FRECR	0x01e8
#define ETH_TSFRCR	0x01ec
#define ETH_TLFRCR	0x01f0
#define ETH_RFCR	0x01f4
#define ETH_MAFCR	0x01f8
#define ETH_RTRATE	0x01fc

#define eth_write(reg, data) write32(ETH_BASE_ADDR + ETH_##reg, data)
#define eth_read(reg) read32(ETH_BASE_ADDR + ETH_##reg)

struct tx_desc {
	union {
		u32 v;
		struct {
			u32 tfs:12,
			    rsv:14,
			    twbi:1,
			    tfe:1,
			    tfp:2,
			    tdle:1,
			    tact:1;
		} s;
	} td0;
	union {
		u32 v;
		struct {
			u32 rsv:16,
			    tdl:16;
		} s;
	} td1;
	union {
		u32 v;
		struct {
			u32 tba;
		} s;
	} td2;
	u32 padding[13]; /* aligned cache line size */
} __attribute__((__packed__));

struct rx_desc {
	union {
		u32 v;
		struct {
			u32 rcs:16,
			    rfs:10,
			    rcse:1,
			    rfe:1,
			    rfp:2,
			    rdle:1,
			    ract:1;
		} s;
	} rd0;
	union {
		u32 v;
		struct {
			u32 rdl:16,
			    rbl:16;
		} s;
	} rd1;
	union {
		u32 v;
		struct {
			u32 rba;
		} s;
	} rd2;
	u32 padding[13];
} __attribute__((__packed__));

#define nr_tx_desc (8)
#define nr_rx_desc (8)
#define desc_buf_size (2048)
static volatile struct tx_desc *tx_desc = (struct tx_desc *)0x40b00000;
static volatile struct rx_desc *rx_desc = (struct rx_desc *)0x40b20000;
static volatile u8 *tx_buf = (u8 *)0x40b40000;
static volatile u8 *rx_buf = (u8 *)0x40b80000;
static volatile struct tx_desc *tx_desc_cur = (struct tx_desc *)0x40b00000;
static volatile struct rx_desc *rx_desc_cur = (struct rx_desc *)0x40b20000;

static void dump(u32 addr, int len)
{
#if defined(DEBUG_DUMP)
	u8 *ptr = (u8 *)addr;
	int i;

	for (i=0; i<len; i++) {
		dev_dbg_r("%b ", *ptr++);
		if (i%16==15)
			dev_dbg_r("\n");
	}
	if (i%16!=16)
		dev_dbg_r("\n");
#endif
}

static u8 __mii_read_bit(void)
{
	u32 pir;

	eth_write(PIR, PIR_MDC);
	udelay(1);
	eth_write(PIR, 0);
	udelay(2);
	pir = eth_read(PIR);
	eth_write(PIR, PIR_MDC);
	udelay(1);

	return pir & PIR_MDI;
}

static void __mii_write_bit(u32 level)
{
	u32 mdo = level ? PIR_MDO : 0;

	eth_write(PIR, mdo | PIR_MMD | PIR_MDC);
	udelay(1);
	eth_write(PIR, mdo | PIR_MMD);
	udelay(2);
	eth_write(PIR, mdo | PIR_MMD | PIR_MDC);
	udelay(1);
}

static u16 mdio_read(struct net_device *dev, addr_t addr, addr_t reg, u16 *val)
{
	u16 _val = 0;
	int i;

	/* Preamble */
	for (i=0; i<32; i++)
		__mii_write_bit(1);

	/* Start of Frame */
	__mii_write_bit(0);
	__mii_write_bit(1);

	/* OP Code: Read */
	__mii_write_bit(1);
	__mii_write_bit(0);

	/* PHY Address */
	for (i=4; i>=0; i--)
		__mii_write_bit(addr & (1 << i));

	/* Register Address */
	for (i=4; i>=0; i--)
		__mii_write_bit(reg & (1 << i));

	/* Turn Around */
	__mii_read_bit();
	__mii_read_bit();

	/* Data */
	for (i=15; i>=0; i--)
		_val |= ((__mii_read_bit() ? 1 : 0) << i);

	if (val)
		*val = _val;

	return 0;
}

static int mdio_write(struct net_device *dev, addr_t addr, addr_t reg, u16 val)
{
	int i;

	/* Preamble */
	for (i=0; i<32; i++)
		__mii_write_bit(1);

	/* Start of Frame */
	__mii_write_bit(0);
	__mii_write_bit(1);

	/* OP Code: Write */
	__mii_write_bit(0);
	__mii_write_bit(1);

	/* PHY Address */
	for (i=4; i>=0; i--)
		__mii_write_bit(addr & (1 << i));

	/* Register Address */
	for (i=4; i>=0; i--)
		__mii_write_bit(reg & (1 << i));

	/* Turn Around */
	__mii_write_bit(0);
	__mii_write_bit(0);

	/* Data */
	for (i=15; i>=0; i--)
		__mii_write_bit(val & (1 << i));

	return 0;
}

static void rcm2_enable_phy(struct net_device *dev)
{
	trace();
}

static void rcm2_disable_phy(struct net_device *dev)
{
	trace();
}

static void rcm2_phy_reset(struct net_device *dev)
{
	u16 val;

	/* Hardware reset using nRST for SMSC LAN8710/LAN8720 */
	gpio_set_value(PIN_ETH_MAGIC, 0);
	udelay(100);
	gpio_set_value(PIN_ETH_MAGIC, 1);
	udelay(1);

	/* Software reset */
	mdio_write(dev, 0x00, 0x00, 0x8000);
	do {
		mdio_read(dev, 0, 0, &val);
	} while (val & 0x8000);
}

static int rcm2_phy_init(struct net_device *dev)
{
	int timeout;
	u16 val;
	u32 id;
	u16 id1, id2;

	trace();

	rcm2_phy_reset(dev);

	mdio_read(dev, 0, 2, &id1);
	mdio_read(dev, 0, 3, &id2);
	id = (id1 << 16 | id2);
	dev_dbg("PHYID = %p, REV = %d\n", id >> 4, id & 0x0f);

	/* starting auto-negotiation and waiting for completion */
	mdio_write(dev, 0x00, 0x00, 0x1200);
	timeout = 5000;
	while (--timeout) {
		mdio_read(dev, 0x00, 0x01, &val);
		if (val & 0x0020)
			break;
		mdelay(1);
		dev_vdbg("Auto-negotiation Satus: %p\n", val);
	}
	if (timeout <= 0) {
		dev_dbg("failed to Auto-negotiation: %p\n", val);
		return -H_EIO;
	}
	dev_dbg("Auto-negotiation Completed: %p\n", val);

	/* waiting for link-up */
	mdelay(300);

	mdio_read(dev, 0x00, 0x01, &val);
	if (!(val & 0x04)) {
		dev_dbg_r("PHY: Link is down!\n");
		return -H_EIO;
	}

#if defined(DEBUG)
	{
		u16 adv, lpa;
		char *speed, *duplex;

		mdio_read(dev, 0x00, 4, &adv);
		mdio_read(dev, 0x00, 5, &lpa);
		lpa &= adv;
		if (lpa & 0x0100) {
			speed = "100BASE-TX";
			duplex = "Full";
		} else if (lpa & 0x0080) {
			speed = "100BASE-TX";
			duplex = "Half";
		} else if (lpa & 0x0040) {
			speed = "10BASE-T";
			duplex = "Full";
		} else if (lpa & 0x0020) {
			speed = "10BASE-T";
			duplex = "Half";
		} else {
			speed = "unknown";
			duplex = "unknown";
		}
		dev_dbg_r("PHY: Link is up! %s/%s\n", speed, duplex);
	}
#endif

	return 0;
}

static void rcm2_module_reset(void)
{
	trace();

	eth_write(EDMR, eth_read(EDMR) | EDMR_SRST);
	udelay(3000);
	eth_write(EDMR, eth_read(EDMR) & ~EDMR_SRST);
}

static void rcm2_mac_reset(void)
{
	trace();

	/* RMII mode */
	write32(0xee70026c, read16(0xee70026c) | 0x0001);
}

static void rcm2_setup_tx_desc(struct net_device *dev)
{
	int i;

	trace();

	for(i = 0; i < nr_tx_desc; i++) {
		memset((void *)&tx_desc[i], 0, sizeof(struct tx_desc));
		tx_desc[i].td0.v = 0;
		tx_desc[i].td1.v = 0;
		tx_desc[i].td2.s.tba = (u32)(tx_buf + (desc_buf_size * i));
	}

	/* Mark the end of the descriptors */
	tx_desc[nr_tx_desc - 1].td0.s.tdle = 1;

	/* Point the controller to the tx descriptor list.  */
	eth_write(TDLAR, (u32)tx_desc);

	tx_desc_cur = tx_desc;
}

static void rcm2_setup_rx_desc(struct net_device *dev)
{
	int i;

	trace();

	for (i=0; i<nr_rx_desc; i++) {
		memset((void *)&rx_desc[i], 0, sizeof(struct rx_desc));
		rx_desc[i].rd0.s.ract = 1;
		rx_desc[i].rd1.s.rbl = desc_buf_size;
		rx_desc[i].rd2.s.rba = (u32)(rx_buf + (desc_buf_size * i));
	}

	/* Mark the end of the descriptors */
	rx_desc[nr_rx_desc - 1].rd0.s.rdle = 1;


	/* Point the controller to the rx descriptor list */
	eth_write(RDLAR, (u32)rx_desc);

	rx_desc_cur = rx_desc;
}

static int rcm2_eth_send(struct net_device *dev, struct net_packet *pkt)
{
	volatile struct tx_desc *desc;
	int len;
	u8 *buf;
	int to = 100000; /* micro seconds */

	trace();

	if (tx_desc_cur == &tx_desc[nr_tx_desc])
		tx_desc_cur = tx_desc;
	desc = tx_desc_cur;

	buf = (u8 *)desc->td2.s.tba;
	memcpy(buf, pkt->eth, ETH_FRAME_LEN);
	len = ETH_FRAME_LEN;
	memcpy(buf + len, pkt->f2.any, pkt->f2len);
	len += pkt->f2len;
	memcpy(buf + len, pkt->f3.any, pkt->f3len);
	len += pkt->f3len;
	memcpy(buf + len, pkt->f4, pkt->f4len);
	len += pkt->f4len;

	/* if (len < 60) */
	/* 	len = 60; */

	desc->td1.s.tdl = len;

	desc->td0.s.tfs = 0;
	desc->td0.s.rsv = 0;
	desc->td0.s.tfe = 0;
	desc->td0.s.tfp = 3;
	desc->td0.s.twbi = 0;
	desc->td0.s.tact = 1;

	/* Restart the transmitter if disabled */
	if (!(eth_read(EDTRR) & EDTRR_TRNS))
		eth_write(EDTRR, EDTRR_TRNS);

	/* Wait until packet is transmitted */
	while (--to) {
		if (!(desc->td0.s.tact)) {
			break;
		}
		udelay(1);
	}

	if (to <= 0)
		return -H_ETIMEDOUT;

	if (desc->td0.s.tact || desc->td0.s.tfe) {
		dev_dbg("Tx error: TD0:%p, EESR=%p,EDTRR=%p\n",
			desc->td0.v, eth_read(EESR),
			eth_read(EDTRR));
		desc->td0.s.tact = 0;
		tx_desc_cur++;
		return -H_EAGAIN;
	}

	tx_desc_cur++;

	return 0;
}

static int rcm2_eth_recv(struct net_device *dev, struct net_packet *pkt,
			 int *timeout)
{
	volatile struct rx_desc *desc;
	int to = *timeout * 1000; /* micro seconds */

	trace();

	if (rx_desc_cur == &rx_desc[nr_rx_desc])
		rx_desc_cur = rx_desc;
	desc = rx_desc_cur;

	while (--to) {
		if (!desc->rd0.s.ract)
			break;
		udelay(1);
	}

	if (to <= 0)
		return -H_ETIMEDOUT;

	if (!desc->rd0.s.ract) {
		if (!(desc->rd0.s.rfp & 1) ||
		    (desc->rd0.s.rfe && (desc->rd0.s.rfs & 0x23f))) {
			dev_dbg("Rx error: RD0:%p\n", desc->rd0.v);
			desc->rd0.s.rcs = 0;
			desc->rd0.s.rfs = 0;
			desc->rd0.s.rcse = 0;
			desc->rd0.s.rfe = 0;
			desc->rd0.s.rfp = 0;
			desc->rd1.s.rdl = 0;
			desc->rd0.s.ract = 1;

			rx_desc_cur++;

			return -H_EAGAIN;
		}

		dump(desc->rd2.s.rba, desc->rd1.s.rdl);
		memcpy(pkt->raw_data_addr, (void *)desc->rd2.s.rba,
		       desc->rd1.s.rdl);
		pkt->raw_data_len = desc->rd1.s.rdl;

		desc->rd0.s.rcs = 0;
		desc->rd0.s.rfs = 0;
		desc->rd0.s.rcse = 0;
		desc->rd0.s.rfe = 0;
		desc->rd0.s.rfp = 0;
		desc->rd1.s.rdl = 0;
		desc->rd0.s.ract = 1;

		/* Restart the receiver if disabled */
		if (!(eth_read(EDRRR) & EDRRR_R))
			eth_write(EDRRR, EDRRR_R);

		rx_desc_cur++;

		return 0;
	}
	dev_dbg("Not Received!\n");

	return -H_EIO;
}

static void rcm2_set_mac(struct net_device *dev, unsigned char *mac)
{
	u32 mac_val;
	mac_val = (mac[0] << 24 |
		   mac[1] << 16 |
		   mac[2] << 8 |
		   mac[3] << 0);
	eth_write(MAHR, mac_val);

	mac_val = (mac[4] << 8 |
		   mac[5] << 0);
	eth_write(MALR, mac_val);
}

static void rcm2_eth_supply_clock(void)
{
	u32 sr, cr;

	sr = read32(CPG_MSTPSR8);
	if (sr & CPG_MSTP8_ETHER) {
		cr = read32(CPG_SMSTPCR8);
		write32(CPG_SMSTPCR8, cr & ~CPG_MSTP8_ETHER);
	}
}

static int rcm2_eth_init(struct net_device *dev)
{
	u32 ecmr;
	u16 adv, lpa;
	int ret;

	trace();

	gpio_direction_output(PIN_ETH_MAGIC, 0);

	if (!is_valid_mac(dev->eth_mac))
		get_mac_address(dev->eth_mac);

	/* module enable */
	rcm2_eth_supply_clock();

	rcm2_module_reset();
	rcm2_mac_reset();

	/* setup tx/rx desc */
	rcm2_setup_tx_desc(dev);
	rcm2_setup_rx_desc(dev);

	/* Configure e-dmac registers */
	eth_write(EDMR, (eth_read(EDMR) & ~EMDR_DESC_R) |
		  EMDR_DESC_64 | EDMR_EL);

	eth_write(EESIPR, 0);
	eth_write(TRSCER, 0);
	eth_write(TFTR, 0);
	eth_write(FDR, FIFO_SIZE_T | FIFO_SIZE_R);
	eth_write(RMCR, RMCR_RST);

	eth_write(FCFTR, FIFO_F_D_RFF | FIFO_F_D_RFD);

	/* Configure e-mac registers */
	eth_write(ECSIPR, 0);

	/* setup emac */
	rcm2_set_mac(dev, dev->eth_mac);

	/* frame length upper limit */
	eth_write(RFLR, RFLR_RFL_MIN);

	/* Initialize phy */
	ret = rcm2_phy_init(dev);
	if (ret) {
		dev_info_r("Link is down\n");
		return ret;
	}

	/* setup speed/duplex */
	ecmr = 0;
	mdio_read(dev, 0x00, 4, &adv);
	mdio_read(dev, 0x00, 5, &lpa);
	lpa &= adv;
	if (lpa & 0x0100) {
		/* 100BASE-TX/Full */
		ecmr |= ECMR_RTM;
		ecmr |= ECMR_DM;
	} else if (lpa & 0x0080) {
		/* 100BASE-TX/Half */
		ecmr |= ECMR_RTM;
	} else if (lpa & 0x0040) {
		/* 10BASE-T/Full */
		ecmr |= ECMR_DM;
	} else {
		/* 10BASE-T/Half */
		;
	}
	eth_write(ECMR, ecmr | ECMR_CHG_DM | ECMR_RE | ECMR_TE);

	/* receive ready */
	eth_write(EDRRR, EDRRR_R);

	return 0;
}

static int rcm2_eth_shutdown(struct net_device *dev)
{
	trace();

	gpio_set_value(PIN_ETH_MAGIC, 0);
	udelay(100);
	gpio_set_value(PIN_ETH_MAGIC, 1);
	udelay(1);

	rcm2_module_reset();
	rcm2_mac_reset();

	rcm2_set_mac(dev, dev->eth_mac);

	eth_write(EDRRR, ~EDRRR_R);

	return 0;
}

struct net_device rcm2_eth = {
	.init = rcm2_eth_init,
	.exit = rcm2_eth_shutdown,

	.set_mac = rcm2_set_mac,
	.enable_phy = rcm2_enable_phy,
	.disable_phy = rcm2_disable_phy,
	.send = rcm2_eth_send,
	.recv = rcm2_eth_recv,
};
