/*
 * Copyright (c) 2014 Atmark Techno, Inc.  All Rights Reserved.
 *  Tetsuhisa KOSEKI <koseki@atmark-techno.com>
 */

#include <autoconf.h>
#include <hermit.h>
#include <buffer.h>
#include <io.h>
#include <herrno.h>
#include <scan.h>
#include <flash.h>
#include <string.h>
#include <delay.h>
#include <spi.h>

/* Common commands */
#define CMD_READ_ID		0x9f

#define CMD_READ_ARRAY_SLOW	0x03
#define CMD_READ_ARRAY_FAST	0x0b
#define CMD_READ_ARRAY_QUAD	0x6c

#define CMD_WRITE_STATUS	0x01
#define CMD_PAGE_PROGRAM	0x02
#define CMD_WRITE_DISABLE	0x04
#define CMD_READ_STATUS		0x05
#define CMD_WRITE_ENABLE	0x06
#define CMD_ERASE_4K		0x20
#define CMD_QUAD4_PAGE_PROGRAM	0x34
#define CMD_READ_CONFIG		0x35
#define CMD_ERASE_32K		0x52
#define CMD_ERASE_64K		0xd8
#define CMD_ERASE_CHIP		0xc7
#define CMD_ERASE_256K		0xdc

/* Spansion commands */
#define CMD_READ_STATUS		0x05
#define CMD_READ_STATUS2	0x35
#define CMD_READ_STATUS3	0x33
#define CMD_FAST_READ_QUAD_OUTPUT	0x6b

#define CMD_DUMMY		0x00

/* Common status */
#define STATUS_WIP		0x01

#define SPI_TIMEOUT		(3 * 1000 * 1000) /* 3(s) */
#define SPI_PAGE_PROG_TIMEOUT	(2 * 1000 * 1000) /* 2(s) */
#define SPI_PAGE_ERASE_TIMEOUT	(5 * 1000 * 1000) /* 5(s) */

struct spi_flash_eb {
	u32 device_id;
	eraseblock_t erase_block;
	u8 (*read_cmd)(u32 addr, void *data, size_t size);
	int (*program_cmd)(const u32 data, const u32 addr, const u32 size);
	int (*erase_cmd)(u32 addr);
	int (*device_init)(void);
};

static void spi_flash_addr(u32 addr, u8 *cmd)
{
	/* cmd[0] is actual command */
	cmd[1] = addr >> 16;
	cmd[2] = addr >> 8;
	cmd[3] = addr >> 0;
}

static void spi_flash_addr4(u32 addr, u8 *cmd)
{
	/* cmd[0] is actual command */
	cmd[1] = addr >> 24;
	cmd[2] = addr >> 16;
	cmd[3] = addr >> 8;
	cmd[4] = addr >> 0;
}

static int spi_flash_cmd_read_id(u32 *id)
{
	int ret;
	u8 cmd;
	u8 data[3];

	cmd = CMD_READ_ID;
	*id = 0;

	ret = spi_xfer_normal(sizeof(cmd), &cmd, NULL, SPI_XFER_BEGIN);
	if (ret)
		return -1;

	ret = spi_xfer_normal(sizeof(data), NULL, &data, SPI_XFER_END);
	if (ret)
		return -1;

	*id = data[2];
	*id |= data[1] << 8;
	*id |= data[0] << 16;

	return 0;
}

static int spi_flash_cmd_write_enable(void)
{
	u8 cmd;

	cmd = CMD_WRITE_ENABLE;

	return spi_xfer_normal(sizeof(cmd), &cmd, NULL,
			       SPI_XFER_BEGIN | SPI_XFER_END);
}

static int spi_flash_cmd_read_status(u8 *status)
{
	int ret;
	u8 cmd;

	cmd = CMD_READ_STATUS;

	ret = spi_xfer_normal(sizeof(cmd), &cmd, NULL, SPI_XFER_BEGIN);
	if (ret)
		return -1;

	ret = spi_xfer_normal(1, NULL, status, SPI_XFER_END);
	if (ret)
		return -1;

	return 0;
}

static int spi_flash_cmd_wait_ready(u32 timeout)
{
	int ret;
	u8 status;

	spi_start_timer(timeout);

	while (spi_check_timer()) {

		ret = spi_flash_cmd_read_status(&status);
		if (ret)
			break;
		else if ((status & STATUS_WIP) == 0) {
			spi_stop_timer();
			return 0;
		}

	}

	spi_stop_timer();

	return -1;
}

static u8 __maybe_unused
flash_spi_read_fast(u32 addr, void *data, size_t size)
{
	u8 cmd[5];
	int ret;

	cmd[0] = CMD_READ_ARRAY_FAST;
	spi_flash_addr(addr, cmd);
	cmd[4] = CMD_DUMMY;

	ret = spi_xfer_normal(sizeof(cmd), cmd, NULL, SPI_XFER_BEGIN);
	if (ret)
		return ret;

	return spi_xfer_normal(size, NULL, data, SPI_XFER_END);
}

static u8 __maybe_unused
flash_spi_read_quad(u32 addr, void *data, size_t size)
{
	u8 cmd[6];

	cmd[0] = CMD_READ_ARRAY_QUAD;
	spi_flash_addr4(addr, cmd);
	cmd[5] = CMD_DUMMY;

	return spi_xfer_quad(sizeof(cmd), cmd, size, NULL, data,
			     SPI_XFER_BEGIN | SPI_XFER_END);
}

static u8 __maybe_unused
flash_spi_read_fast_quad_output(u32 addr, void *data, size_t size)
{
	u8 cmd[5];

	cmd[0] = CMD_FAST_READ_QUAD_OUTPUT;
	spi_flash_addr(addr, cmd);
	cmd[4] = CMD_DUMMY;

	return spi_xfer_quad(sizeof(cmd), cmd, size, NULL, data,
			     SPI_XFER_BEGIN | SPI_XFER_END);
}

static int __maybe_unused
spi_page_program_normal(const u32 data, const u32 addr, const u32 size)
{
	u8 cmd[4];
	int ret;

	cmd[0] = CMD_PAGE_PROGRAM;
	spi_flash_addr(addr, cmd);

	ret = spi_flash_cmd_write_enable();
	if (ret < 0)
		return ret;

	ret = spi_xfer_normal(sizeof(cmd), cmd, NULL, SPI_XFER_BEGIN);
	if (ret)
		return ret;

	ret = spi_xfer_normal(size, (u8 *)data, NULL, SPI_XFER_END);
	if (ret)
		return ret;

	return spi_flash_cmd_wait_ready(SPI_PAGE_PROG_TIMEOUT);
}

static int __maybe_unused
spi_page_program_quad(const u32 data, const u32 addr, const u32 size)
{
	u8 cmd[5];
	int ret;

	cmd[0] = CMD_QUAD4_PAGE_PROGRAM;
	spi_flash_addr4(addr, cmd);

	ret = spi_flash_cmd_write_enable();
	if (ret < 0)
		return ret;

	ret = spi_xfer_quad(sizeof(cmd), cmd, size, (u8 *)data, NULL,
			    SPI_XFER_BEGIN | SPI_XFER_END);
	if (ret < 0)
		return ret;

	return spi_flash_cmd_wait_ready(SPI_PAGE_PROG_TIMEOUT);
}

static int __maybe_unused
flash_spi_erase_normal(u32 addr)
{
	u8 cmd[4];
	int ret;

	cmd[0] = CMD_ERASE_64K;

	spi_flash_addr(addr, cmd);

	ret = spi_flash_cmd_write_enable();
	if (ret)
		return ret;

	ret = spi_xfer_normal(sizeof(cmd), cmd, NULL,
			      SPI_XFER_BEGIN | SPI_XFER_END);
	if (ret)
		return ret;

	ret = spi_flash_cmd_wait_ready(SPI_PAGE_ERASE_TIMEOUT);

	return ret;
}

static int __maybe_unused
spi_flash_spansion_cmd_read_status2(u8 *status)
{
	int ret;
	u8 cmd;

	cmd = CMD_READ_STATUS2;

	ret = spi_xfer_normal(sizeof(cmd), &cmd, NULL, SPI_XFER_BEGIN);
	if (ret)
		return -1;

	ret = spi_xfer_normal(1, NULL, status, SPI_XFER_END);
	if (ret)
		return -1;

	return 0;
}

static int __maybe_unused
spi_flash_spansion_cmd_read_status3(u8 *status)
{
	int ret;
	u8 cmd;

	cmd = CMD_READ_STATUS3;

	ret = spi_xfer_normal(sizeof(cmd), &cmd, NULL, SPI_XFER_BEGIN);
	if (ret)
		return -1;

	ret = spi_xfer_normal(1, NULL, status, SPI_XFER_END);
	if (ret)
		return -1;

	return 0;
}

static int __maybe_unused
spi_flash_spansion_cmd_write_status(const u8 reg[3])
{
	u8 cmd;
	int ret;

	cmd = CMD_WRITE_STATUS;

	ret = spi_flash_cmd_write_enable();
	if (ret < 0)
		return ret;

	ret = spi_xfer_normal(sizeof(cmd), &cmd, NULL, SPI_XFER_BEGIN);
	if (ret)
		return ret;

	ret = spi_xfer_normal(sizeof(reg), reg, NULL, SPI_XFER_END);
	if (ret)
		return ret;

	return spi_flash_cmd_wait_ready(SPI_PAGE_PROG_TIMEOUT);
}

static int __maybe_unused
spi_flash_spansion_cmd_quad_enable(void)
{
	int ret;
	u8 reg[3];

	ret = spi_flash_cmd_read_status(&reg[0]);
	if (ret)
		return -1;

	ret = spi_flash_spansion_cmd_read_status2(&reg[1]);
	if (ret)
		return -1;

	ret = spi_flash_spansion_cmd_read_status3(&reg[2]);
	if (ret)
		return -1;

	/* quad enable */
	reg[1] |= 0x02;

	return spi_flash_spansion_cmd_write_status(reg);
}

static struct spi_flash_eb *p_spi_flash_eb = NULL;

static struct spi_flash_eb spi_flash_table[] = {
	{	/* S25FL512S */
		.device_id = 0x010220,
		.erase_block = { 256, 1024 },
#if defined(CONFIG_FLASH_SPI_USE_QUAD)
		.read_cmd = flash_spi_read_quad,
		.program_cmd = spi_page_program_quad,
#else
		.read_cmd = flash_spi_read_fast,
		.program_cmd = spi_page_program_normal,
#endif
		.erase_cmd = flash_spi_erase_normal,
	},
	{	/* S25FL164K */
		.device_id = 0x014017,
		.erase_block = { 128, 256 },
#if defined(CONFIG_FLASH_SPI_USE_QUAD)
		.read_cmd = flash_spi_read_fast_quad_output,
		.program_cmd = spi_page_program_normal,
		.device_init = spi_flash_spansion_cmd_quad_enable,
#else
		.read_cmd = flash_spi_read_fast,
		.program_cmd = spi_page_program_normal,
#endif
		.erase_cmd = flash_spi_erase_normal,
	},
	{	/* N25Q64 */
		.device_id = 0x20ba17,
		.erase_block = { 128, 256 },
#if defined(CONFIG_FLASH_SPI_USE_QUAD)
		.read_cmd = flash_spi_read_fast_quad_output,
#else
		.read_cmd = flash_spi_read_fast,
#endif
		.program_cmd = spi_page_program_normal,
		.erase_cmd = flash_spi_erase_normal,
	},
	{	/* M25P64 */
		.device_id = 0x202017,
		.erase_block = { 128, 256 },
		.read_cmd = flash_spi_read_fast,
		.program_cmd = spi_page_program_normal,
		.erase_cmd = flash_spi_erase_normal,
	},
};

int flash_spi_initialize(const u32 base_addr)
{
	int i;
	int ret;
	u32 device_id;

	ret = spi_initialize();
	if (ret)
		return -1;

	device_id = flash_spi_get_id(base_addr);

	for (i = 0; i < ARRAY_SIZE(spi_flash_table); i++) {
		if (spi_flash_table[i].device_id == device_id) {
			p_spi_flash_eb = &spi_flash_table[i];
			if (p_spi_flash_eb->device_init)
				p_spi_flash_eb->device_init();
			return 0;
		}
	}

	hprintf("Error: Unknown SPI Flash Device ID = 0x%x\n", device_id);

	return -1;
}

u8 flash_spi_read_byte(u32 from)
{
	u8 data;

	data = 0x00;

	p_spi_flash_eb->read_cmd(from, &data, 1);

	return data;
}

u32 flash_spi_copy_to_dram(const u32 from, const u32 to, const u32 size)
{
	p_spi_flash_eb->read_cmd(from, (u8 *)to, size);

	return size;
}

#define PROGRAM_BUFFER_SIZE (256)
#define MIN(a, b) (((a) < (b)) ? (a) : (b))
int flash_spi_program(const u32 from, const u32 to, const u32 size)
{
	int ret;
	u32 actual;
	u32 program_size;
	u32 byte_addr;

	if(from & UNALIGNED_MASK) return -H_EALIGN;
	if(to & UNALIGNED_MASK) return -H_EALIGN;

	byte_addr = to % PROGRAM_BUFFER_SIZE;
	for(actual = 0; actual < size; actual += program_size){
		program_size = MIN(size - actual,
				   PROGRAM_BUFFER_SIZE - byte_addr);

		ret = p_spi_flash_eb->program_cmd(from + actual,
						  to + actual, program_size);
		if (ret)
			return ret;

		byte_addr = 0;
	}

	return 0;
}

int flash_spi_erase(const u32 erase_addr)
{
	u32 addr;

	/* word align */
	addr = (erase_addr & ~UNALIGNED_MASK);

	return p_spi_flash_eb->erase_cmd(addr);
}

u32 flash_spi_get_id(const u32 base_addr)
{
	u32 id;

	spi_flash_cmd_read_id(&id);

	return id;
}

int flash_spi_get_size(const u32 base_addr)
{
	int ret;
	int flash_size;

	if (!p_spi_flash_eb)
		return -1;

	flash_size = p_spi_flash_eb->erase_block.num *
		p_spi_flash_eb->erase_block.size * 256;

	ret = 0;
	while (flash_size > 1) {
		flash_size /= 2;
		ret++;
	}

	return ret;
}

int flash_spi_create_eraseblock_table(const u32 base_addr, flash_eb *eb)
{
	if (!p_spi_flash_eb)
		return -1;

	eb->nr_eraseblock = 1;
	eb->base_addr = base_addr;
	eb->eraseblock[0] = p_spi_flash_eb->erase_block;

	return 0;
}
